from abstract_utilities import get_any_value,SingletonMeta
import logging,os,json
from abstract_pandas import pd
from abstract_apis import getRequest
from PIL import Image
import io
import requests
from ..imageUtils import *
from ..dbConfig import get_meta_data_by_mint
def get_image_vars(uri):
    response = getRequest(url=uri,data={})
    return response
def get_image_data(image_url):
    try:
        response = getRequest(url=uri,data={},result='content')
        image = Image.open(io.BytesIO(response.content))
        image.thumbnail((200, 200))
        bio = io.BytesIO()
        image.save(bio, format="PNG")
        return bio.getvalue()
    except:
        pass
metaDataLowerKeys = {
    'image': 'image',
    'uri': 'uri',
    'supply': 'supply',
    'symbol': 'symbol',
    'mintauthority': 'mintAuthority',
    'freezeauthority': 'freezeAuthority',
    'twitter': 'twitter',
    'website': 'website',
    'ismutable': 'isMutable',
    'creators': 'creators',
    'updateauthority': 'updateAuthority',
    'createdon': 'createdOn',
    'primarysalehappened': 'createdOn',
    'description': 'description'
}
metaDataTypeKeys = {
    'image': str,
    'uri': str,
    'supply': int,
    'name':str,
    'symbol': str,
    'mintAuthority': str,
    'freezeAuthority': str,
    'twitter': str,
    'website': str,
    'isMutable': bool,
    'creators': str,
    'updateAuthority': str,
    'createdOn': int,
    'primarySaleHappened': bool,
    'description': str
}
metaDataCheckKeys = {
    'image': True,
    'uri': True,
    'supply': True,
    'symbol': True,
    'name':True,
    'mintAuthority': True,
    'freezeAuthority': True,
    'twitter': True,
    'website': True,
    'isMutable': True,
    'creators': True,
    'updateAuthority': True,
    'createdOn': True,
    'primarySaleHappened': True,
    'description': True
}
metaKeys = list(metaDataTypeKeys.keys())

def make_insert(key):
    key = key.replace(' ', '_').upper()
    return f"-{key}-"
def make_insert_bool(key):
    key = key.replace(' ', '_').upper()
    return f"-{key}_BOOL-"
def make_check_bool(key):
    key = key.replace(' ', '_').upper()
    return f"-{key}_CHECK-"
def deKeyKey(key):
    return key.lower()[1:-1].replace('_',' ')
def getBool(key, value):
    return isinstance(value, metaDataTypeKeys.get(key))
class metaDataManager(metaclass=SingletonMeta):
    def __init__(self):
        self.allMetaData = {}
        self.metaDataCheckKeys=metaDataCheckKeys
        self.metaDataTypeKeys = metaDataTypeKeys
    def changeTally(self,event,values):
        lower_key = deKeyKey(event)
        regularMetakey = metaDataLowerKeys.get(lower_key)
        self.metaDataCheckKeys[regularMetakey] = values[event]
    def processMetaData(self,mint,imageData=True):
        if not self.allMetaData.get(mint):
            metaData = get_meta_data_by_mint(mint)
            self.get_meta_vars(metaData,imageData)
            self.allMetaData[mint] = {"processed":self.datas_js,'bools':self.bool_js}
        return self.allMetaData[mint]
    def filter_meta_data(self,mint,filterMetaChecks={}):
        metaData_js = self.processMetaData(mint)
        bools = metaData_js['bools']
        self.allMetaData[mint]['filtered']=True
        for key,check in self.metaDataCheckKeys.items():
            if check and not bools.get(key):
                self.allMetaData[mint]['filtered']=False
                break
        return self.allMetaData[mint]['filtered']
    def get_meta_vars(self,metaData,imageData=True):
        self.datas_js = {}
        self.bool_js = {}
        self.metaData = self.get_uri(metaData,imageData)
        for key in metaKeys:
            self.get_from_any_meta(key,metaData)
    def get_from_any_meta(self,key,metaData):
        values = get_any_value(metaData, key) or None
        value = values[0] if values and isinstance(values, list) else values
        self.bool_js[key] =getBool(key, value)
        self.datas_js[key] = value
        return value
    def get_uri(self,metaData,imageData=True):
        key = 'uri'
        self.get_from_any_meta(key,metaData)
        
        if self.bool_js[key] and imageData:
           try:
                metaData[0].update(get_image_vars(self.datas_js[key]) or {})
           except:
                pass
        return metaData
            




from dateutil import parser
from functools import lru_cache
import itertools,math
from ..dbConfig import *
from .transactionProcessor import TransactionProcessor
from abstract_utilities import make_list
from abstract_security import *
from datetime import datetime
from statistics import mean
from ..managers.metaDataManager import *
from .poolMetrics import *
metaDataMgr = metaDataManager()
tables="""token_metadata
transactions
owneraccounts
token_data
wallet_account_assignments
account_bonding_curves
transaction_signatures
metadata
useraddresses
logdata
pairs
getsignaturesforaddress
"""
def get_env_path():
    return '/home/development/finalDatabase/.env'


def get_key_value(key, path=None):
    path = path or get_env_path()
    return get_env_value(key=key, path=path)




def get_signature_placement(obj,num):
    int((num-len(str(obj)))/2)
    newStr = ''
    for i in range(int((num-len(str(obj)))/2)):
       newStr+=' ' 
    return f"{newStr}{obj}{newStr}"
def get_pairs():
    tab = '\t'
    all_pairs =get_all_pairs()
    
    for pair in reversed(all_pairs):
        processor = TransactionProcessor(pair)
        user_wallets = processor.user_wallets
        profits = processor.profits
        user_wallets = processor.user_wallets
        pair['transactions'] = processor.clean_txns(processor.all_txns)
        for wallet,values in profits.items():
            print(f"wallet: {wallet}")
            unique_txns = processor.get_sorted_unique_txns(values['txns'])
        
            print(f"{tab}{tab}txn_no{tab}{tab}{get_signature_placement('signature',94)}{tab}{tab}{get_signature_placement('amount',14)}{tab}{tab}{get_signature_placement('volume',14)}")
            for txn in unique_txns:
                sig = ""
                if txn['signature'][-1] ==  ' ':
                    sig = f"{tab}"
                print(f"{tab}{tab}{txn['txn_number']}{tab}{tab}{txn['signature']}__{sig}{tab}{tab}{get_signature_placement(txn['sol_amount'],14)}{tab}{tab}{txn['volume']}")
        input(processor.ranking())    




import logging
# Configure logging
logging.basicConfig(level=logging.INFO)  # Set level to INFO or higher
import traceback,requests,psycopg2,json
from abstract_utilities import make_list,is_number
from abstract_apis import *
from abstract_security import get_env_value
keysConvertJS = {'log_id':'logdata', 'meta_id':'metadata','txn_id':'transactions','pair_id':'pairs'}
from .asyncUtils import call_solcatcher_db

def call_solcatcher_db(endpoint,*args,**kwargs):
    response = call_solcatcher_db(endpoint,*args,**kwargs)
    return get_response(response)
def get_column_names(tableName,schema='public'):
    return call_solcatcher_db('get_column_names',tableName=tableName,schema=schema)
def aggregate_rows(query, values=None, errorMsg='Error Fetching Rows'):
    return call_solcatcher_db('aggregate_rows',query=query, values=values, errorMsg=errorMsg)
def fetch_any_combo(columnNames=None, tableName=None, searchColumn=None, searchValue=None, anyValue=False, zipit=True):
    if tableName and columnNames==None:
        columnNames = get_column_names(tableName).get('columns')
    return call_solcatcher_db('fetch_any_combo',columnNames=columnNames, tableName=tableName, searchColumn=searchColumn, searchValue=searchValue, anyValue=anyValue, zipit=zipit)
def getZipRows(tableName, rows, schema='public'):
    return call_solcatcher_db('getZipRows',tableName=tableName,rows=rows,schema=schema)
def get_signatures(address,before=None):
    return call_solcatcher_db('call-signatures',address=address,before=before)
def get_all_wallet_assignments():
    return call_solcatcher_db('all-wallet-assignments')
def update_signatures(address,before=None):
    return call_solcatcher_db('update-signatures',address=address,before=before)
def assign_wallet(address):
    return call_solcatcher_db('assign-wallet',address=address)
def get_existing_signatures(address):
    return call_solcatcher_db('get-signatures',address=address)
def get_main_data_columns(main_columns,main_data):
    return call_solcatcher_db('get-get_main_data_columns',main_columns=main_columns,main_data=main_data)
def generate_chart(txn_history):
    return call_solcatcher_db('get-generate_chart',txn_history=txn_history)
def insertSignature(address):
    return call_solcatcher_db('get-generate_chart',txn_history=txn_history)
def get_assigned_account(address):
    query = "SELECT assigned_account FROM wallet_account_assignments WHERE address = %s;"
    assigned_account = aggregate_rows(query,(address,))
    if assigned_account:
        assigned_account = assigned_account[0][0]
    return assigned_account
def get_all_addresses_in_assigned_account(assigned_account):
    query = "SELECT address FROM wallet_account_assignments WHERE assigned_account = %s;"
    return aggregate_rows(query,(assigned_account,))
def get_all_pairs_for_addresses_in_assigned_account(addresses):
    query = "SELECT id FROM wallet_account_assignments WHERE user_address = ANY(%s);"
    return aggregate_rows(query,(addresses,))
def get_all_mints_for_addresses_in_assigned_account(ids):
    query = "SELECT mint FROM pairs WHERE id = ANY(%s);"
    return aggregate_rows(query,(ids,))
def get_all_assigned_accounts_for_address(address):
    assigned_account = get_assigned_account(address)
    if not assigned_account:
        assigned_account = update_signatures(address)
        if assigned_account:
           assigned_account = assigned_account.get('assigned_account')
    addresses = get_all_addresses_in_assigned_account(assigned_account)
    return addresses
def get_all_pairs_for_address(address):
    addresses = get_all_assigned_accounts_for_address(address)
    pair_ids = get_all_pairs_for_addresses_in_assigned_account(addresses)
    return pair_ids
def get_all_mints_for_address(address):
    pair_ids = get_all_pairs_for_address(address)
    mints = get_all_mints_for_addresses_in_assigned_account(pair_ids)
    return mints
def search_table_for_id(tableName,value):
    allPairRows = fetch_any_combo(columnNames='*', tableName=tableName,searchColumn='id',searchValue=value)
    return allPairRows
def get_all_txns_for_pair(*args,**kwargs):
    pair_id = get_pair_id(*args,**kwargs)
    allPairRows = fetch_any_combo(columnNames='*', tableName='transactions',searchColumn='pair_id',searchValue=pair_id)
    return allPairRows
def get_pair_id(arg=None,id=None,mint=None,signature=None,pair=None):
    if arg:
        if isinstance(arg,dict):
            pair = arg
        elif is_number(arg):
            id = arg
        else:
            mint=arg
            signature=arg
        
    if pair:
        searchColumn='id'
        id = pair.get('id')
    if mint or signature:
        if mint:
            searchColumn='mint'
            searchValue=mint
        else:
            searchColumn='signature'
            searchValue=signature
        rows= fetch_any_combo(columnNames='*', tableName='transactions', searchColumn=searchColumn,searchValue=searchValue)
        if rows:
            id = rows.get('id')    
    return id
def get_pair_by(**kwargs):
    for key,value in kwargs.items():
        if key=='pair':
            return pair
        rows= fetch_any_combo(columnNames='*', tableName='transactions', searchColumn=key,searchValue=value)
        if rows:
            if isinstance(rows,list):
                rows = rows[0]
            return rows
def get_pair_data(**kwargs):
    pair = get_pair_by(**kwargs)
    pair_id = pair.get('id')
    row = fetch_any_combo(columnNames='*', tableName='pairs',searchColumn='id',searchValue=pair_id)
    pair_column = get_column_data(row)
    transactions = get_all_txns_for_pair(pair_id)
    pair_column['transactions']=get_all_txns_for_pair(pair_id)
    return pair_column
def get_column_data(row,exclude=['log_id']):
    new_js = {}
    exclude = make_list(exclude)
    for key,value in row[0].items():
        if key not in exclude:
            newKey = keysConvertJS.get(key)
            if newKey:
                newValue = search_table_for_id(newKey,value)
                if newValue:
                    key = newKey
                    value = newValue
            new_js[key] =value
    return new_js

def fetch_filtered_transactions(sol_amount, operator="<", timestamp=None, timestamp_operator=None):
    """
    Fetch filtered transactions based on Solana amount and optional timestamp filtering.

    :param sol_amount: Solana amount to filter transactions by.
    :param operator: Comparison operator for sol_amount ('>', '<=', etc.). Default is '>'.
    :param timestamp: Optional timestamp for filtering transactions.
    :param timestamp_operator: Operator for timestamp comparison ('>', '<=', etc.).
    :return: List of filtered transactions.
    """
    # Validate operators
    valid_operators = [">", "<=", "=", "<", ">="]
    if operator not in valid_operators:
        raise ValueError(f"Invalid operator for sol_amount: {operator}. Must be one of {valid_operators}")
    if timestamp_operator and timestamp_operator not in valid_operators:
        raise ValueError(f"Invalid operator for timestamp: {timestamp_operator}. Must be one of {valid_operators}")

    # Start SQL query
    query = f"""
    SELECT 
        t.id AS transaction_id,
        t.signature AS transaction_signature,
        t.tcns AS transaction_details,
        p.id AS pair_id,
        p.signature AS genesis_signature,
        p.mint AS mint_address
    FROM 
        transactions t
    JOIN 
        pairs p 
    ON 
        t.pair_id = p.id
    WHERE 
        p.signature IS NOT NULL
    AND 
        t.program_id = p.program_id
    AND 
        EXISTS (
            SELECT 1
            FROM jsonb_array_elements(t.tcns) AS elem
            WHERE 
                (elem ->> 'sol_amount')::numeric {operator} %s
    """

    # Add timestamp filter if provided
    params = [sol_amount]
    if timestamp and timestamp_operator:
        query += f" AND (elem ->> 'timestamp')::bigint {timestamp_operator} %s"
        params.append(int(timestamp))  # Ensure timestamp is an integer

    # Close EXISTS clause
    query += ")"

    # Final ordering clause
    query += " ORDER BY t.updated_at DESC;"

    # Execute query
    return aggregate_rows(query, params)

def fetch_filtered_transactions(sol_amount, operator="<", timestamp=None, timestamp_operator=None):
    """
    Fetch filtered transactions based on Solana amount and optional timestamp filtering.

    :param sol_amount: Solana amount to filter transactions by.
    :param operator: Comparison operator for sol_amount ('>', '<=', etc.). Default is '>'.
    :param timestamp: Optional timestamp for filtering transactions.
    :param timestamp_operator: Operator for timestamp comparison ('>', '<=', etc.).
    :return: List of filtered transactions.
    """
    # Validate operators
    valid_operators = [">", "<=", "=", "<", ">="]
    if operator not in valid_operators:
        raise ValueError(f"Invalid operator for sol_amount: {operator}. Must be one of {valid_operators}")
    if timestamp_operator and timestamp_operator not in valid_operators:
        raise ValueError(f"Invalid operator for timestamp: {timestamp_operator}. Must be one of {valid_operators}")

    # Start SQL query
    query = f"""
    SELECT 
        t.id AS transaction_id,
        t.signature AS transaction_signature,
        t.tcns AS transaction_details,
        p.id AS pair_id,
        p.signature AS genesis_signature,
        p.mint AS mint_address
    FROM 
        transactions t
    JOIN 
        pairs p 
    ON 
        t.pair_id = p.id
    WHERE 
        p.signature IS NOT NULL
    AND 
        t.program_id = p.program_id
    AND 
        EXISTS (
            SELECT 1
            FROM jsonb_array_elements(t.tcns) AS elem
            WHERE 
                (elem ->> 'sol_amount')::numeric {operator} %s
    """

    # Add timestamp filter if provided
    params = [sol_amount]
    if timestamp and timestamp_operator:
        query += f" AND (elem ->> 'timestamp')::bigint {timestamp_operator} %s"
        params.append(int(timestamp))  # Ensure timestamp is an integer

    # Close EXISTS clause
    query += ")"

    # Final ordering clause
    query += " ORDER BY t.updated_at DESC;"

    # Execute query
    return aggregate_rows(query, params)

import time
def fetch_filtered_transactions_paginated(
    sol_amount, 
    operator="<", 
    timestamp=None, 
    timestamp_operator=None, 
    limit=50, 
    offset=0
):
    query = f"""
    SELECT 
        t.id AS transaction_id,
        t.signature AS transaction_signature,
        t.tcns AS transaction_details,
        p.id AS pair_id,
        p.signature AS genesis_signature,
        p.mint AS mint_address
    FROM 
        transactions t
    JOIN 
        pairs p 
    ON 
        t.pair_id = p.id
    WHERE 
        p.signature IS NOT NULL
    AND 
        t.program_id = p.program_id
    AND 
        EXISTS (
            SELECT 1
            FROM jsonb_array_elements(t.tcns) AS elem
            WHERE (elem ->> 'sol_amount')::numeric {operator} %s
    """

    params = [sol_amount]
    if timestamp and timestamp_operator:
        query += f" AND (elem ->> 'timestamp')::bigint {timestamp_operator} %s"
        params.append(int(timestamp))

    query += f") ORDER BY t.updated_at DESC LIMIT %s OFFSET %s;"
    params.extend([limit, offset])

    return aggregate_rows(query, params)

def get_time(minutes=0, hours=0, days=0, weeks=0, years=0):
    """Calculate a UNIX timestamp for the given time offset."""
    mins = 60
    hr = 60 * mins
    day = 24 * hr
    week = 7 * day
    year = 365 * day
    timeStamp = time.time() - ((mins * minutes) + (hr * hours) + (day * days) + (week * weeks) + (year * years))
    return int(timeStamp)  # Return integer timestamp
def ge_time(minutes=0, hours=0, days=0, weeks=0, years=0):
    """Calculate a UNIX timestamp for the given time offset."""
    mins = 60
    hr = 60 * mins
    day = 24 * hr
    week = 7 * day
    year = 365 * day
    timeStamp = time.time() - ((mins * minutes) + (hr * hours) + (day * days) + (week * weeks) + (year * years))
    return int(timeStamp)  # Return integer timestamp
